from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib import ticker
from typing_extensions import Literal

from .._conversion import db
from .._helper import verify_scalar
from ._utility import stem


def real_or_complex_plot(
    x: npt.NDArray,
    y: npt.NDArray,
    ax: plt.Axes,
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    label = kwargs.pop("label", None)
    if np.iscomplexobj(y):
        if label is None:
            label, label2 = "real", "imag"
        else:
            label, label2 = label + " (real)", label + " (imag)"

        if diff == "color":
            ax.plot(x, y.real, label=label, **kwargs)
            ax.plot(x, y.imag, label=label2, **kwargs)
        elif diff == "line":
            (real,) = ax.plot(x, y.real, "-", label=label, **kwargs)
            kwargs.pop("color", None)
            ax.plot(x, y.imag, "--", color=real.get_color(), label=label2, **kwargs)
        else:
            raise ValueError(f"Argument 'diff' must be 'color' or 'line', not {diff}.")
    else:
        ax.plot(x, y, label=label, **kwargs)

    if label:
        ax.legend()


def standard_plot(
    x: npt.NDArray,
    y: npt.NDArray,
    ax: plt.Axes,
    type: Literal["plot", "stem"] = "plot",
    y_axis: Literal["complex", "mag", "mag^2", "db"] = "mag",
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    if y_axis == "complex":
        pass
    elif y_axis == "mag":
        y = np.abs(y)
    elif y_axis == "mag^2":
        y = np.abs(y) ** 2
    elif y_axis == "db":
        y = db(np.abs(y) ** 2)
    else:
        raise ValueError(f"Argument 'y_axis' must be 'complex', 'mag', 'mag^2', or 'db', not {y_axis!r}.")

    label = kwargs.pop("label", None)

    if type == "plot":
        if np.iscomplexobj(y):
            if label is None:
                label, label2 = "real", "imag"
            else:
                label, label2 = label + " (real)", label + " (imag)"

            if diff == "color":
                ax.plot(x, y.real, label=label, **kwargs)
                ax.plot(x, y.imag, label=label2, **kwargs)
            elif diff == "line":
                (real,) = ax.plot(x, y.real, "-", label=label, **kwargs)
                kwargs.pop("color", None)
                ax.plot(x, y.imag, "--", color=real.get_color(), label=label2, **kwargs)
            else:
                raise ValueError(f"Argument 'diff' must be 'color' or 'line', not {diff}.")
        else:
            ax.plot(x, y, label=label, **kwargs)

    elif type == "stem":
        stem(x, y, ax=ax, label=label, **kwargs)

    else:
        raise ValueError(f"Argument 'type' must be 'plot' or 'stem', not {type!r}.")

    if label:
        ax.legend()


def verify_sample_rate(sample_rate: float | None, default=1.0):
    if sample_rate is None:
        sample_rate_provided = False
        sample_rate = default
    else:
        sample_rate_provided = True
        verify_scalar(sample_rate, float=True, positive=True)

    return sample_rate, sample_rate_provided


def sample_units(min: float, max: float) -> tuple[str, float]:
    """
    Determines the appropriate time units to use for a given sample array.
    """
    max_sample = np.max([np.abs(max), np.abs(min)])

    if max_sample > 1e12:
        scalar = 1e-12
        units = "Tsamples"
    elif max_sample > 1e9:
        scalar = 1e-9
        units = "Gsamples"
    elif max_sample > 1e6:
        scalar = 1e-6
        units = "Msamples"
    elif max_sample > 1e3:
        scalar = 1e-3
        units = "ksamples"
    else:
        scalar = 1
        units = "samples"

    return units, scalar


def time_units(min: float, max: float) -> tuple[str, float]:
    """
    Determines the appropriate time units to use for a given time array.
    """
    max_time = np.max([np.abs(max), np.abs(min)])

    if max_time > 1:
        scalar = 1
        units = "s"
    elif max_time > 1e-3:
        scalar = 1e3
        units = "ms"
    elif max_time > 1e-6:
        scalar = 1e6
        units = "μs"
    elif max_time > 1e-9:
        scalar = 1e9
        units = "ns"
    elif max_time > 1e-12:
        scalar = 1e12
        units = "ps"
    else:
        scalar = 1e15
        units = "fs"

    return units, scalar


def freq_units(min: float, max: float) -> tuple[str, float]:
    """
    Determines the appropriate frequency units to use for a given frequency array.
    """
    max_freq = np.max([np.abs(max), np.abs(min)])

    if max_freq > 1e12:
        scalar = 1e-12
        units = "THz"
    elif max_freq > 1e9:
        scalar = 1e-9
        units = "GHz"
    elif max_freq > 1e6:
        scalar = 1e-6
        units = "MHz"
    elif max_freq > 1e3:
        scalar = 1e-3
        units = "kHz"
    elif max_freq >= 0.5:
        scalar = 1
        units = "Hz"
    elif max_freq > 1e-3:
        scalar = 1e3
        units = "mHz"
    else:
        scalar = 1e6
        units = "μHz"

    return units, scalar


def smart_scaled_formatter(val, pos, scalar):
    """
    Format the tick label based on the scale factor and the value.

    The goal is to provide a label with the appropriate number of decimal places.
    """
    scaled = val * scalar

    if abs(scaled - round(scaled)) < 1e-6:
        label = f"{int(round(scaled))}"
    elif abs(1e1 * scaled - round(1e1 * scaled)) < 1e-6:
        label = f"{scaled:.1f}"
    elif abs(1e2 * scaled - round(1e2 * scaled)) < 1e-6:
        label = f"{scaled:.2f}"
    elif abs(1e3 * scaled - round(1e3 * scaled)) < 1e-6:
        label = f"{scaled:.3f}"
    elif abs(1e4 * scaled - round(1e4 * scaled)) < 1e-6:
        label = f"{scaled:.4f}"
    elif abs(1e5 * scaled - round(1e5 * scaled)) < 1e-6:
        label = f"{scaled:.5f}"
    else:
        label = f"{scaled:.6f}"

    return label


def time_x_axis(ax: plt.Axes, sample_rate_provided: bool) -> str:
    """
    Format the x-axis of a plot to display time in seconds.
    """
    if sample_rate_provided:
        units, scalar = time_units(*ax.get_xlim())
        ax.set_xlabel(f"Time ({units}), $t$")
    else:
        units, scalar = sample_units(*ax.get_xlim())
        if scalar == 1:
            ax.set_xlabel("Sample, $n$")
            ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
            ax.xaxis.set_minor_locator(ticker.MaxNLocator(integer=True))
        else:
            ax.set_xlabel(f"Sample ({units}), $n$")

    formatter = ticker.FuncFormatter(lambda pos, val: smart_scaled_formatter(pos, val, scalar))
    ax.xaxis.set_major_formatter(formatter)

    return units


def time_y_axis(ax: plt.Axes, sample_rate_provided: bool) -> str:
    """
    Format the y-axis of a plot to display time in seconds.
    """
    if sample_rate_provided:
        units, scalar = time_units(*ax.get_ylim())
        ax.set_ylabel(f"Time ({units}), $t$")
    else:
        units, scalar = sample_units(*ax.get_ylim())
        if scalar == 1:
            ax.set_ylabel("Sample, $n$")
            ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))
            ax.yaxis.set_minor_locator(ticker.MaxNLocator(integer=True))
        else:
            ax.set_ylabel(f"Sample ({units}), $n$")

    formatter = ticker.FuncFormatter(lambda pos, val: smart_scaled_formatter(pos, val, scalar))
    ax.yaxis.set_major_formatter(formatter)

    return units


def freq_x_axis(ax: plt.Axes, sample_rate_provided: bool, bins: bool = False) -> str:
    """
    Format the x-axis of a plot to display frequency in Hz.
    """
    if bins:
        units, scalar = "", 1
        ax.set_xlabel("DFT bin index, $k$")
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        ax.xaxis.set_minor_locator(ticker.MaxNLocator(integer=True))
    elif sample_rate_provided:
        units, scalar = freq_units(*ax.get_xlim())
        ax.set_xlabel(f"Frequency ({units}), $f$")
    else:
        units, scalar = "", 1
        ax.set_xlabel("Normalized frequency, $f/f_s$")

    formatter = ticker.FuncFormatter(lambda pos, val: smart_scaled_formatter(pos, val, scalar))
    ax.xaxis.set_major_formatter(formatter)

    return units


def freq_y_axis(ax: plt.Axes, sample_rate_provided: bool, bins: bool = False) -> str:
    """
    Format the y-axis of a plot to display frequency in Hz.
    """
    if bins:
        units, scalar = "", 1
        ax.set_ylabel("DFT bin index, $k$")
        ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        ax.yaxis.set_minor_locator(ticker.MaxNLocator(integer=True))
    elif sample_rate_provided:
        units, scalar = freq_units(*ax.get_ylim())
        ax.set_ylabel(f"Frequency ({units}), $f$")
    else:
        units, scalar = "", 1
        ax.set_ylabel("Normalized frequency, $f/f_s$")

    formatter = ticker.FuncFormatter(lambda pos, val: smart_scaled_formatter(pos, val, scalar))
    ax.yaxis.set_major_formatter(formatter)

    return units


def min_ylim(y: npt.NDArray, separation: float, sample_rate: float):
    ymin, ymax = plt.gca().get_ylim()
    if ymax - ymin < separation / 4:
        # Find the mean of the signal rounded to the nearest sample
        mean = int(round(np.nanmean(y) / sample_rate))
        ymin = mean - separation / 2
        ymax = mean + separation / 2
        plt.gca().set_ylim(ymin, ymax)



import numpy as np
import pytest

import sdr


@pytest.mark.parametrize("error", [1e-9, 1e-6, 1e-3, 1e-1])
def test_real_passband(error):
    sample_rate = 2e6
    freq = 100e3
    duration = 1000e-6
    x = sdr.sinusoid(duration, freq, sample_rate=sample_rate, complex=False)

    y = sdr.clock_error(x, error)

    error = -error / (1 + error)
    z = sdr.clock_error(y, error)

    # Ignore the first and last 5 sample due to filter effects. The signals aren't perfectly identical also due
    # to filter effects. Use abolute tolerance to verify the signals are spinning at different rates.
    np.testing.assert_allclose(x[5 : z.size - 5], z[5:-5], atol=0.01)


@pytest.mark.parametrize("error", [1e-9, 1e-6, 1e-3, 1e-1])
def test_complex_passband(error):
    sample_rate = 2e6
    center_freq = 1e6
    duration = 1000e-6
    x = sdr.sinusoid(duration, 0, sample_rate=sample_rate)

    y = sdr.clock_error(x, error, 0, center_freq, sample_rate=sample_rate)

    error = -error / (1 + error)
    z = sdr.clock_error(y, error, 0, center_freq, sample_rate=sample_rate)

    # Ignore the first and last 5 sample due to filter effects. The signals aren't perfectly identical also due
    # to filter effects. Use abolute tolerance to verify the signals are spinning at different rates.
    np.testing.assert_allclose(x[5 : z.size - 5], z[5:-5], atol=0.01)



"""
A module containing various probability functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.special

from ._helper import convert_output, export, verify_arraylike, verify_scalar


@export
def Q(x: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Computes the CCDF of the standard normal distribution $\mathcal{N}(0, 1)$..

    The complementary cumulative distribution function (CCDF) $Q(x)$ is the probability that a random variable
    exceeds a given value.

    Arguments:
        x: The real-valued input $x$.

    Returns:
        The probability $p$ that $x$ is exceeded.

    See Also:
        sdr.Qinv

    Examples:
        .. ipython:: python

            sdr.Q(1)
            sdr.Qinv(0.158655)

    Group:
        probability
    """
    x = verify_arraylike(x, float=True)

    p = scipy.special.erfc(x / np.sqrt(2)) / 2

    return convert_output(p)


@export
def Qinv(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Computes the inverse CCDF of the standard normal distribution $\mathcal{N}(0, 1)$.

    The inverse complementary cumulative distribution function (CCDF) $Q^{-1}(p)$ is the value that is exceeded
    with a given probability.

    Arguments:
        p: The probability $p$ of exceeding the returned value $x$.

    Returns:
        The real-valued $x$ that is exceeded with probability $p$.

    See Also:
        sdr.Q

    Examples:
        .. ipython:: python

            sdr.Qinv(0.158655)
            sdr.Q(1)

    Group:
        probability
    """
    p = verify_arraylike(p, float=True)

    x = np.sqrt(2) * scipy.special.erfcinv(2 * p)

    return convert_output(x)


@export
def add_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the sum of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the random variable $X$.
        Y: The distribution of the random variable $Y$.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = X + Y$.

    Notes:
        Given two independent random variables $X$ and $Y$ with PDFs $f_X(x)$ and $f_Y(y)$, we compute the PDF of
        $Z = X + Y$ as follows.

        The PDF of $Z$, denoted $f_Z(z)$, can be obtained using the convolution formula for independent random
        variables.

        $$f_Z(z) = \int_{-\infty}^\infty f_X(x) f_Y(z - x) \, dx$$

    Examples:
        Compute the distribution of the sum of normal and Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            Y = scipy.stats.rayleigh(loc=1, scale=2)
            x = np.linspace(-4, 10, 1_001)

            @savefig sdr_add_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.add_rvs(X, Y).pdf(x), label="$X + Y$"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X + Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of normal and Rayleigh random variables");

        Compute the distribution of the sum of Rayleigh and Rician random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            Y = scipy.stats.rice(3)
            x = np.linspace(0, 12, 1_001)

            @savefig sdr_add_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.add_rvs(X, Y).pdf(x), label="$X + Y$"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X + Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of Rayleigh and Rician random variables");

        Compute the distribution of the sum of Rician and Chi-squared random variables.

        .. ipython:: python

            X = scipy.stats.rice(3)
            Y = scipy.stats.chi2(3)
            x = np.linspace(0, 20, 1_001)

            @savefig sdr_add_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.add_rvs(X, Y).pdf(x), label="$X + Y$"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X + Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of Rician and Chi-squared random variables");

    Group:
        independent-rvs
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x_min, x_max = _x_range(X, p)
    y_min, y_max = _x_range(Y, p)
    dx = (x_max - x_min) / 1_000
    dy = (y_max - y_min) / 1_000
    dz = np.min([dx, dy])  # Use the smaller delta x -- must use the same dx for both distributions
    x = np.arange(x_min, x_max, dz)
    y = np.arange(y_min, y_max, dz)

    # Compute the PDF of each distribution
    f_X = X.pdf(x)
    f_Y = Y.pdf(y)

    # The PDF of the sum of two independent random variables is the convolution of the PDF of the two distributions
    f_Z = np.convolve(f_X, f_Y, mode="full") * dz

    # Determine the x axis for the output convolution
    z = np.arange(f_Z.size) * dz + x[0] + y[0]

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def subtract_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the difference of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the random variable $X$.
        Y: The distribution of the random variable $Y$.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the difference $Z = X - Y$.

    Notes:
        Given two independent random variables $X$ and $Y$ with PDFs $f_X(x)$ and $f_Y(y)$, we compute the PDF of
        $Z = X - Y$ as follows.

        The PDF of $Z$, denoted $f_Z(z)$, can be obtained using the convolution formula for independent random
        variables. For the difference $Z = X - Y$, the PDF of $Y$ is flipped.

        $$f_Z(z) = \int_{-\infty}^\infty f_X(x) f_Y(x - z) \, dx$$

    Examples:
        Compute the distribution of the difference of normal and Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=5, scale=0.5)
            Y = scipy.stats.rayleigh(loc=1, scale=2)
            x = np.linspace(-5, 10, 1_001)

            @savefig sdr_subtract_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.subtract_rvs(X, Y).pdf(x), label="$X - Y$"); \
            plt.hist(X.rvs(100_000) - Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X - Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Difference of normal and Rayleigh random variables");

        Compute the distribution of the difference of Rayleigh and Rician random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            Y = scipy.stats.rice(3)
            x = np.linspace(-10, 10, 1_001)

            @savefig sdr_subtract_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.subtract_rvs(X, Y).pdf(x), label="$X - Y$"); \
            plt.hist(X.rvs(100_000) - Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X - Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Difference of Rayleigh and Rician random variables");

        Compute the distribution of the difference of Rician and Chi-squared random variables.

        .. ipython:: python

            X = scipy.stats.rice(3)
            Y = scipy.stats.chi2(3)
            x = np.linspace(-10, 10, 1_001)

            @savefig sdr_subtract_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.subtract_rvs(X, Y).pdf(x), label="$X - Y$"); \
            plt.hist(X.rvs(100_000) - Y.rvs(100_000), bins=101, density=True, histtype="step", label="$X - Y$ empirical"); \
            plt.legend(); \
            plt.xlim(-10, 10); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Difference of Rician and Chi-squared random variables");

    Group:
        independent-rvs
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x_min, x_max = _x_range(X, p)
    y_min, y_max = _x_range(Y, p)  # Compute the bounds for Y
    y_min, y_max = -y_max, -y_min  # Compute the bounds for -Y
    dx = (x_max - x_min) / 1_000
    dy = (y_max - y_min) / 1_000
    dz = np.min([dx, dy])  # Use the smaller delta x -- must use the same dx for both distributions
    x = np.arange(x_min, x_max, dz)
    y = np.arange(y_min, y_max, dz)

    # Compute the PDF of each distribution
    f_X = X.pdf(x)
    f_Y = Y.pdf(-y)  # Compute the PDF of -Y

    # The PDF of the sum of two independent random variables is the convolution of the PDF of the two distributions
    f_Z = np.convolve(f_X, f_Y, mode="full") * dz

    # Determine the x axis for the output convolution
    z = np.arange(f_Z.size) * dz + x[0] + y[0]

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def multiply_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    z: npt.ArrayLike | None = None,
    p: float = 1e-10,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the product of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the random variable $X$.
        Y: The distribution of the random variable $Y$.
        z: The $z$ values at which to evaluate the PDF of $Z$. If None, the $z$ values are determined based on `p`.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the product $Z = X \cdot Y$.

    Notes:
        Given two independent random variables $X$ and $Y$ with PDFs $f_X(x)$ and $f_Y(y)$, and CDFs $F_X(x)$ and
        $F_Y(y)$, we compute the PDF of $Z = X \cdot Y$ as follows.

        The PDF of $Z$, denoted $f_Z(z)$, can be derived using the joint distribution of $X$ and $Y$. Since
        $Z = X \cdot Y$, we express the relationship between $x$, $y$, and $z$ and use a transformation approach.

        Let $Z = X \cdot Y$. The PDF $f_Z(z)$ is given by

        $$f_Z(z) = \int_{-\infty}^\infty \frac{1}{\left| w \right|} f_X\left(\frac{z}{w}\right) f_Y(w) \, dw$$

        The Jacobian adjustment for this transformation contributes the factor $\frac{1}{\left| w \right|}$.

    Examples:
        Compute the distribution of the product of two normal random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            Y = scipy.stats.norm(loc=2, scale=1.5)
            x = np.linspace(-15, 10, 1_001)

            @savefig sdr_multiply_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.multiply_rvs(X, Y).pdf(x), label=r"$X \cdot Y$"); \
            plt.hist(X.rvs(100_000) * Y.rvs(100_000), bins=101, density=True, histtype="step", label=r"$X \cdot Y$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Product of two normal random variables");

    Group:
        independent-rvs
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if z is None:
        # Determine the z axis of each distribution such that the probability of exceeding the z axis, on either side,
        # is p.
        x_min, x_max = _x_range(X, np.sqrt(p))
        y_min, y_max = _x_range(Y, np.sqrt(p))
        bounds = np.array([x_min * y_min, x_min * y_max, x_max * y_min, x_max * y_max])
        z_min = np.min(bounds)
        z_max = np.max(bounds)
        z = np.linspace(z_min, z_max, 1_001)
    else:
        z = verify_arraylike(z, float=True, atleast_1d=True, ndim=1)
        z = np.sort(z)
    dz = np.mean(np.diff(z))

    def integrand(y: float, z: float) -> float:
        return 1 / np.abs(y) * X.pdf(z / y) * Y.pdf(y)

    f_Z = np.zeros_like(z)
    for i, zi in enumerate(z):
        f_Z[i] = scipy.integrate.quad(integrand, -np.inf, np.inf, args=(zi,))[0]

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def min_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the minimum of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the random variable $X$.
        Y: The distribution of the random variable $Y$.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the minimum $Z = \min(X, Y)$.

    Notes:
        Given two independent random variables $X$ and $Y$ with PDFs $f_X(x)$ and $f_Y(y)$, and CDFs $F_X(x)$ and
        $F_Y(y)$, we compute the PDF of $Z = \min(X, Y)$ as follows.

        The CDF of $Z$, denoted $F_Z(z)$, is $F_Z(z) = P(Z \leq z)$. Since $Z = \min(X, Y)$, the event $Z \leq z$
        occurs if either $X \leq z$ or $Y \leq z$. Using the complement and independence,

        $$P(Z \leq z) = 1 - P(Z > z) = 1 - P(X > z \text{ and } Y > z)$$
        $$F_Z(z) = 1 - P(X > z) \cdot P(Y > z)$$
        $$F_Z(z) = 1 - (1 - F_X(z)) \cdot (1 - F_Y(z)) .$$

        The PDF of $Z$, denoted $f_Z(z)$, is the derivative of $F_Z(z)$. Therefore, $f_Z(z) = \frac{d}{dz} F_Z(z)$.
        Substituting $F_Z(z) = 1 - (1 - F_X(z)) \cdot (1 - F_Y(z))$ yields

        $$f_Z(z) = \frac{d}{dz} \big(1 - (1 - F_X(z)) \cdot (1 - F_Y(z))\big)$$
        $$f_Z(z) = f_X(z) \cdot (1 - F_Y(z)) + f_Y(z) \cdot (1 - F_X(z)) .$$

        Therefore, the PDF of $Z = \min(X, Y)$ is

        $$f_Z(z) = f_X(z) \cdot (1 - F_Y(z)) + f_Y(z) \cdot (1 - F_X(z))$$

        where $F_X(z)$ and $F_Y(z)$ are the CDFs of the original random variables $X$ and $Y$, and $f_X(z)$ and
        $f_Y(z)$ are the PDFs of $X$ and $Y$.

    Examples:
        Compute the distribution of the minimum of normal and Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=3, scale=0.5)
            Y = scipy.stats.rayleigh(loc=1, scale=2)
            x = np.linspace(0, 10, 1_001)

            @savefig sdr_min_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.min_rvs(X, Y).pdf(x), label=r"$\min(X, Y)$"); \
            plt.hist(np.minimum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\min(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of normal and Rayleigh random variables");

        Compute the distribution of the minimum of Rayleigh and Rician random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=2)
            Y = scipy.stats.rice(2)
            x = np.linspace(0, 8, 1_001)

            @savefig sdr_min_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.min_rvs(X, Y).pdf(x), label=r"$\min(X, Y)$"); \
            plt.hist(np.minimum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\min(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of Rayleigh and Rician random variables");

        Compute the distribution of the minimum of Rician and Chi-squared random variables.

        .. ipython:: python

            X = scipy.stats.rice(3)
            Y = scipy.stats.chi2(3)
            x = np.linspace(0, 20, 1_001)

            @savefig sdr_min_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.min_rvs(X, Y).pdf(x), label=r"$\min(X, Y)$"); \
            plt.hist(np.minimum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\min(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlim(0, 15); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of Rician and Chi-squared random variables");

    Group:
        independent-rvs
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x_min, x_max = _x_range(X, p)
    y_min, y_max = _x_range(Y, p)
    dx = (x_max - x_min) / 1_000
    dy = (y_max - y_min) / 1_000
    dz = np.min([dx, dy])  # Use the smaller delta x -- must use the same dx for both distributions
    z = np.arange(np.min([x_min, y_min]), np.max([x_max, y_max]), dz)

    # Compute the PDF of each distribution
    f_X = X.pdf(z)
    f_Y = Y.pdf(z)

    # Compute the CDF of each distribution
    F_X = X.cdf(z)
    F_Y = Y.cdf(z)

    # The PDF of the min of X and Y
    f_Z = f_X * (1 - F_Y) + f_Y * (1 - F_X)

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def max_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the maximum of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the random variable $X$.
        Y: The distribution of the random variable $Y$.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the maximum $Z = \max(X, Y)$.

    Notes:
        Given two independent random variables $X$ and $Y$ with PDFs $f_X(x)$ and $f_Y(y)$, and CDFs $F_X(x)$ and
        $F_Y(y)$, we compute the PDF of $Z = \max(X, Y)$ as follows.

        The CDF of $Z$, denoted $F_Z(z)$, is $F_Z(z) = P(Z \leq z)$. Since $Z = \max(X, Y)$, the event $Z \leq z$
        occurs if and only if both $X \leq z$ and $Y \leq z$. Using independence,

        $$F_Z(z) = P(X \leq z) \cdot P(Y \leq z) = F_X(z) \cdot F_Y(z) .$$

        The PDF of $Z$, denoted $f_Z(z)$, is the derivative of $F_Z(z)$. Therefore, $f_Z(z) = \frac{d}{dz} F_Z(z)$.
        Substituting $F_Z(z) = F_X(z) \cdot F_Y(z)$ yields

        $$f_Z(z) = \frac{d}{dz} \big(F_X(z) \cdot F_Y(z)\big)$$
        $$f_Z(z) = f_X(z) \cdot F_Y(z) + f_Y(z) \cdot F_X(z) .$$

        Therefore, the PDF of $Z = \max(X, Y)$ is

        $$f_Z(z) = f_X(z) \cdot F_Y(z) + f_Y(z) \cdot F_X(z)$$

        where $F_X(z)$ and $F_Y(z)$ are the CDFs of the original random variables $X$ and $Y$, and $f_X(z)$ and
        $f_Y(z)$ are the PDFs of $X$ and $Y$.

    Examples:
        Compute the distribution of the maximum of normal and Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=3, scale=0.5)
            Y = scipy.stats.rayleigh(loc=1, scale=2)
            x = np.linspace(0, 10, 1_001)

            @savefig sdr_max_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.max_rvs(X, Y).pdf(x), label=r"$\max(X, Y)$"); \
            plt.hist(np.maximum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\max(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of normal and Rayleigh random variables");

        Compute the distribution of the maximum of Rayleigh and Rician random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            Y = scipy.stats.rice(2)
            x = np.linspace(0, 6, 1_001)

            @savefig sdr_max_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.max_rvs(X, Y).pdf(x), label=r"$\max(X, Y)$"); \
            plt.hist(np.maximum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\max(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of Rayleigh and Rician random variables");

        Compute the distribution of the maximum of Rician and Chi-squared random variables.

        .. ipython:: python

            X = scipy.stats.rice(3)
            Y = scipy.stats.chi2(3)
            x = np.linspace(0, 20, 1_001)

            @savefig sdr_max_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, Y.pdf(x), label="$Y$"); \
            plt.plot(x, sdr.max_rvs(X, Y).pdf(x), label=r"$\max(X, Y)$"); \
            plt.hist(np.maximum(X.rvs(100_000), Y.rvs(100_000)), bins=101, density=True, histtype="step", label=r"$\max(X, Y)$ empirical"); \
            plt.legend(); \
            plt.xlim(0, 15); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of Rician and Chi-squared random variables");

    Group:
        independent-rvs
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x_min, x_max = _x_range(X, p)
    y_min, y_max = _x_range(Y, p)
    dx = (x_max - x_min) / 1_000
    dy = (y_max - y_min) / 1_000
    dz = np.min([dx, dy])  # Use the smaller delta x -- must use the same dx for both distributions
    z = np.arange(np.min([x_min, y_min]), np.max([x_max, y_max]), dz)

    # Compute the PDF of each distribution
    f_X = X.pdf(z)
    f_Y = Y.pdf(z)

    # Compute the CDF of each distribution
    F_X = X.cdf(z)
    F_Y = Y.cdf(z)

    # The PDF of the min of X and Y
    f_Z = f_X * F_Y + f_Y * F_X

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def add_iid_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    n_vars: int,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the sum of $n$ i.i.d. random variables $X_i$.

    Arguments:
        X: The distribution of the i.i.d. random variables $X_i$.
        n_vars: The number $n$ of random variables.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = X_1 + X_2 + \dots + X_n$.

    Notes:
        Given a random variable $X$ with PDF $f_X(x)$, we compute the PDF of
        $Z = X_1 + X_2 + \dots + X_n$, where $X_1, X_2, \dots, X_n$ are independent and identically distributed
        (i.i.d.), as follows.

        The PDF of $Z$, denoted $f_Z(z)$, can be obtained by using the convolution formula for independent
        random variables. Specifically, the PDF of the sum of $n$ i.i.d. random variables is given by the $n$-fold
        convolution of the PDF of $X$ with itself.

        For $n = 2$, $Z = X_1 + X_2$. The PDF of $Z$ is

        $$f_Z(z) = \int_{-\infty}^\infty f_X(x) f_X(z - x) \, dx$$

        For $n > 2$, the PDF of $Z = X_1 + X_2 + \dots + X_n$ is computed recursively

        $$f_Z(z) = \int_{-\infty}^\infty f_X(x) f_{Z_{n-1}}(z - x) \, dx$$

        where $f_{Z_{n-1}}(z)$ is the PDF of the sum of $n-1$ random variables.

        For large $n$, the Central Limit Theorem may be used as an approximation. If $X_i$ have mean $\mu$ and
        variance $\sigma^2$, then $Z$ approximately follows $Z \sim \mathcal{N}(n\mu, n\sigma^2)$
        for sufficiently large $n$.

    Examples:
        Compute the distribution of the sum of two normal random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            n_vars = 2
            x = np.linspace(-6, 2, 1_001)

            @savefig sdr_add_iid_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.add_iid_rvs(X, n_vars).pdf(x), label="$X_1 + X_2$"); \
            plt.hist(X.rvs((100_000, n_vars)).sum(axis=1), bins=101, density=True, histtype="step", label="$X_1 + X_2$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of two normal random variables");

        Compute the distribution of the sum of three Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            n_vars = 3
            x = np.linspace(0, 10, 1_001)

            @savefig sdr_add_iid_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.add_iid_rvs(X, n_vars).pdf(x), label="$X_1 + X_2 + X_3$"); \
            plt.hist(X.rvs((100_000, n_vars)).sum(axis=1), bins=101, density=True, histtype="step", label="$X_1 + X_2 + X_3$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of three Rayleigh random variables");

        Compute the distribution of the sum of four Rician random variables.

        .. ipython:: python

            X = scipy.stats.rice(2)
            n_vars = 4
            x = np.linspace(0, 18, 1_001)

            @savefig sdr_add_iid_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.add_iid_rvs(X, n_vars).pdf(x), label="$X_1 + X_2 + X_3 + X_4$"); \
            plt.hist(X.rvs((100_000, n_vars)).sum(axis=1), bins=101, density=True, histtype="step", label="$X_1 + X_2 + X_3 + X_4$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of four Rician random variables");

    Group:
        iid-rvs
    """
    verify_scalar(n_vars, int=True, positive=True)
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if n_vars == 1:
        return X

    # Compute the exact distribution, if possible
    if isinstance(X.dist, scipy.stats.rv_continuous):
        shape, loc, scale = X.dist._parse_args(*X.args, **X.kwds)

        if isinstance(X.dist, type(scipy.stats.norm)):
            # Sum of normals is normal
            mu = loc  # Mean
            sigma = scale  # Standard deviation
            mu_sum = n_vars * mu
            sigma_sum = np.sqrt(n_vars) * sigma
            return scipy.stats.norm(loc=mu_sum, scale=sigma_sum)
        if isinstance(X.dist, type(scipy.stats.expon)):
            # Sum of exponentials is gamma
            lambda_inv = 1 / X.mean()  # Rate parameter lambda
            return scipy.stats.gamma(a=n_vars, scale=1 / lambda_inv)
        if isinstance(X.dist, type(scipy.stats.gamma)):
            # Sum of gammas with the same scale is gamma
            a = shape[0]  # Shape parameter a
            a_sum = a * n_vars
            return scipy.stats.gamma(a=a_sum, scale=scale)
        if isinstance(X.dist, type(scipy.stats.chi2)):
            # Sum of Chi-Squares is Chi-Square
            df = shape[0]
            df_sum = n_vars * df
            return scipy.stats.chi2(df=df_sum)
    # if isinstance(X.dist, scipy.stats.rv_discrete):
    #     shape, loc, scale = X.dist._parse_args(*X.args, **X.kwds)

    #     if isinstance(X.dist, type(scipy.stats.poisson)):
    #         # Sum of Poissons is Poisson
    #         mu = shape[0]  # Rate parameter mu
    #         mu_sum = mu * n_vars
    #         return scipy.stats.poisson(mu=mu_sum)
    #     if isinstance(X.dist, type(scipy.stats.bernoulli)):
    #         # Sum of Bernoullis is Binomial
    #         p = shape[1]  # Probability of success
    #         return scipy.stats.binom(n=n_vars, p=p)
    #     if isinstance(X.dist, type(scipy.stats.geom)):
    #         # Sum of Geometrics is Negative Binomial
    #         p = shape[0]  # Probability of success
    #         return scipy.stats.binom(n=n_vars, p=p)

    # TODO: Add an override that uses the Central Limit Theorem for large values of n

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    z1_min, z1_max = _x_range(X, p)
    z = np.linspace(z1_min, z1_max, 1_001)
    dz = np.mean(np.diff(z))

    # Compute the PDF of the base distribution
    f_X = X.pdf(z)

    # The PDF of the sum of n_vars independent random variables is the convolution of the PDF of the base distribution.
    # This is efficiently computed in the frequency domain by exponentiating the FFT. The FFT must be zero-padded
    # enough that the circular convolutions do not wrap around.
    n_fft = scipy.fft.next_fast_len(f_X.size * n_vars)
    f_X_fft = np.fft.fft(f_X, n_fft)
    f_X_fft = f_X_fft**n_vars
    f_Z = np.fft.ifft(f_X_fft).real
    f_Z /= f_Z.sum() * dz
    z = np.arange(f_Z.size) * dz + z[0] * (n_vars)

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def min_iid_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    n_vars: int,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the minimum of $n$ i.i.d. random variables $X_i$.

    Arguments:
        X: The distribution of the i.i.d. random variables $X_i$.
        n_vars: The number $n$ of random variables.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = \min(X_1, X_2, \dots, X_n)$.

    Notes:
        Given a random variable $X$ with PDF $f_X(x)$ and CDF $F_X(x)$, we compute the PDF of
        $Z = \min(X_1, X_2, \dots, X_n)$, where $X_1, X_2, \dots, X_n$ are independent and identically distributed
        (i.i.d.), as follows.

        The CDF of $Z$, denoted $F_Z(z)$, is $F_Z(z) = P(Z \leq z)$. Since $Z = \min(X_1, X_2, \dots, X_n)$, the
        event $Z \leq z$ occurs if at least one $X_i \leq z$. Using independence,

        $$F_Z(z) = 1 - P(\text{All } X_i > z) = 1 - \prod_{i=1}^n P(X_i > z) = [1 - F_X(z)]^n = 1 - [1 - F_X(z)]^n .$$

        The PDF of $Z$, denoted $f_Z(z)$, is the derivative of $F_Z(z)$. Therefore, $f_Z(z) = \frac{d}{dz} F_Z(z)$.
        Substituting $F_Z(z) = 1 - [1 - F_X(z)]^n$ yields $f_Z(z) = n \cdot [1 - F_X(z)]^{n-1} \cdot f_X(z)$.

        Therefore, the PDF of $Z = \min(X_1, X_2, \dots, X_n)$ is

        $$f_Z(z) = n \cdot [1 - F_X(z)]^{n-1} \cdot f_X(z)$$

        where $F_X(z)$ is the CDF of the original random variable $X$, $f_X(z)$ is the PDF of $X$, and $n$ is the
        number of samples.

    Examples:
        Compute the distribution of the minimum of ten normal random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            n_vars = 10
            x = np.linspace(-4, 1, 1_001)

            @savefig sdr_min_iid_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.min_iid_rvs(X, n_vars).pdf(x), label=r"$\min(X_1, X_2)$"); \
            plt.hist(X.rvs((100_000, n_vars)).min(axis=1), bins=101, density=True, histtype="step", label=r"$\min(X_1, X_2)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of 10 normal random variables");

        Compute the distribution of the minimum of ten Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            n_vars = 10
            x = np.linspace(0, 4, 1_001)

            @savefig sdr_min_iid_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.min_iid_rvs(X, n_vars).pdf(x), label="$\\min(X_1, \dots, X_3)$"); \
            plt.hist(X.rvs((100_000, n_vars)).min(axis=1), bins=101, density=True, histtype="step", label="$\\min(X_1, \dots, X_3)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of 10 Rayleigh random variables");

        Compute the distribution of the minimum of 100 Rician random variables.

        .. ipython:: python

            X = scipy.stats.rice(2)
            n_vars = 100
            x = np.linspace(0, 6, 1_001)

            @savefig sdr_min_iid_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.min_iid_rvs(X, n_vars).pdf(x), label=r"$\min(X_1, \dots, X_{100})$"); \
            plt.hist(X.rvs((100_000, n_vars)).min(axis=1), bins=101, density=True, histtype="step", label=r"$\min(X_1, \dots, X_{100})$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Minimum of 100 Rician random variables");

    Group:
        iid-rvs
    """
    verify_scalar(n_vars, int=True, positive=True)
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if n_vars == 1:
        return X

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    z1_min, z1_max = _x_range(X, p)
    z = np.linspace(z1_min, z1_max, 1_001)
    dz = np.mean(np.diff(z))

    # Compute the PDF and CDF of the base distribution
    f_X = X.pdf(z)
    F_X = X.cdf(z)

    f_Z = n_vars * (1 - F_X) ** (n_vars - 1) * f_X

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


@export
def max_iid_rvs(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    n_vars: int,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the maximum of $n$ i.i.d. random variables $X_i$.

    Arguments:
        X: The distribution of the i.i.d. random variables $X_i$.
        n_vars: The number $n$ of random variables.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = \max(X_1, X_2, \dots, X_n)$.

    Notes:
        Given a random variable $X$ with PDF $f_X(x)$ and CDF $F_X(x)$, we compute the PDF of
        $Z = \max(X_1, X_2, \dots, X_n)$, where $X_1, X_2, \dots, X_n$ are independent and identically distributed
        (i.i.d.), as follows.

        The CDF of $Z$, denoted $F_Z(z)$, is $F_Z(z) = P(Z \leq z)$. Since $Z = \max(X_1, X_2, \dots, X_n)$, the
        event $Z \leq z$ occurs if and only if all $X_i \leq z$. Using independence,

        $$F_Z(z) = P(Z \leq z) = \prod_{i=1}^n P(X_i \leq z) = [F_X(z)]^n .$$

        The PDF of $Z$, denoted $f_Z(z)$, is the derivative of $F_Z(z)$. Therefore, $f_Z(z) = \frac{d}{dz} F_Z(z)$.
        Substituting $F_Z(z) = [F_X(z)]^n$ yields $f_Z(z) = n \cdot [F_X(z)]^{n-1} \cdot f_X(z)$.

        Therefore, the PDF of $Z = \max(X_1, X_2, \dots, X_n)$ is

        $$f_Z(z) = n \cdot [F_X(z)]^{n-1} \cdot f_X(z)$$

        where $F_X(z)$ is the CDF of the original random variable $X$, $f_X(z)$ is the PDF of $X$, and $n$ is the
        number of samples.

    Examples:
        Compute the distribution of the maximum of two normal random variables.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            n_vars = 2
            x = np.linspace(-3, 1, 1_001)

            @savefig sdr_max_iid_rvs_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.max_iid_rvs(X, n_vars).pdf(x), label=r"$\max(X_1, X_2)$"); \
            plt.hist(X.rvs((100_000, n_vars)).max(axis=1), bins=101, density=True, histtype="step", label=r"$\max(X_1, X_2)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of two normal random variables");

        Compute the distribution of the maximum of ten Rayleigh random variables.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            n_vars = 10
            x = np.linspace(0, 6, 1_001)

            @savefig sdr_max_iid_rvs_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.max_iid_rvs(X, n_vars).pdf(x), label="$\\max(X_1, \dots, X_3)$"); \
            plt.hist(X.rvs((100_000, n_vars)).max(axis=1), bins=101, density=True, histtype="step", label="$\\max(X_1, \dots, X_3)$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of 10 Rayleigh random variables");

        Compute the distribution of the maximum of 100 Rician random variables.

        .. ipython:: python

            X = scipy.stats.rice(2)
            n_vars = 100
            x = np.linspace(0, 8, 1_001)

            @savefig sdr_max_iid_rvs_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="$X$"); \
            plt.plot(x, sdr.max_iid_rvs(X, n_vars).pdf(x), label=r"$\max(X_1, \dots, X_{100})$"); \
            plt.hist(X.rvs((100_000, n_vars)).max(axis=1), bins=101, density=True, histtype="step", label=r"$\max(X_1, \dots, X_{100})$ empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Maximum of 100 Rician random variables");

    Group:
        iid-rvs
    """
    verify_scalar(n_vars, int=True, positive=True)
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if n_vars == 1:
        return X

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    z1_min, z1_max = _x_range(X, p)
    z = np.linspace(z1_min, z1_max, 1_001)
    dz = np.mean(np.diff(z))

    # Compute the PDF and CDF of the base distribution
    f_X = X.pdf(z)
    F_X = X.cdf(z)

    f_Z = n_vars * F_X ** (n_vars - 1) * f_X

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    z = np.append(z, z[-1] + dz)
    z -= dz / 2

    return scipy.stats.rv_histogram((f_Z, z))


def _x_range(X: scipy.stats.rv_continuous, p: float) -> tuple[float, float]:
    r"""
    Determines the range of x values for a given distribution such that the probability of exceeding the x axis, on
    either side, is p.
    """
    # Need to have these loops because for very small p, sometimes SciPy will return NaN instead of a valid value.
    # The while loops will increase the p value until a valid value is returned.

    pp = p
    while True:
        x_min = X.ppf(pp)
        if not np.isnan(x_min):
            break
        pp *= 10

    pp = p
    while True:
        x_max = X.isf(pp)
        if not np.isnan(x_max):
            break
        pp *= 10

    return x_min, x_max



"""
A module containing functions for simulating various signal impairments.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import linear
from .._farrow import FarrowResampler
from .._helper import (
    convert_output,
    export,
    verify_arraylike,
    verify_not_specified,
    verify_only_one_specified,
    verify_scalar,
)
from .._measurement import average_power


@export
def awgn(
    x: npt.ArrayLike,
    snr: float | None = None,
    noise: float | None = None,
    seed: int | None = None,
) -> npt.NDArray:
    r"""
    Adds additive white Gaussian noise (AWGN) to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which AWGN is added.
        snr: The desired signal-to-noise ratio (SNR) in dB. If specified, the average signal power is measured
            explicitly. It is assumed that $x[n]$ contains signal only. If the signal power is known, the
            desired noise variance can be computed and passed in `noise`. If `snr` is `None`,
            `noise` must be specified.
        noise: The noise power (variance) in linear units. If `noise` is `None`, `snr` must be specified.
        seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.

    Returns:
        The noisy signal $x[n] + w[n]$.

    Notes:
        The signal-to-noise ratio (SNR) is defined as

        $$
        \text{SNR} = \frac{P_{\text{signal,avg}}}{P_{\text{noise}}}
        = \frac{\frac{1}{N} \sum_{n=0}^{N-1} \left| x[n] \right|^2}{\sigma^2} ,
        $$

        where $\sigma^2$ is the noise variance. The output signal, with the specified SNR, is $y[n] = x[n] + w[n]$.

        For real signals:
        $$w \sim \mathcal{N}(0, \sigma^2)$$

        For complex signals:
        $$w \sim \mathcal{CN}(0, \sigma^2) = \mathcal{N}(0, \sigma^2 / 2) + j\mathcal{N}(0, \sigma^2 / 2)$$

    Examples:
        Create a real sinusoid and set its $S/N$ to 10 dB.

        .. ipython:: python

            x = np.sin(2 * np.pi * 5 * np.arange(100) / 100); \
            y = sdr.awgn(x, snr=10)

            @savefig sdr_awgn_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="$x[n]$"); \
            sdr.plot.time_domain(y, label="$y[n]$"); \
            plt.title("Input signal $x[n]$ and noisy output signal $y[n]$ with 10 dB SNR");

        Create a QPSK reference signal and set its $E_s/N_0$ to 10 dB. When the signal has 1 sample per symbol,
        $E_s/N_0$ is equivalent to the discrete-time $S/N$.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, psk.order, 1_000); \
            x = psk.map_symbols(s); \
            y = sdr.awgn(x, snr=10)

            @savefig sdr_awgn_2.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title(f"QPSK constellations for $x[n]$ with $\infty$ dB $E_s/N_0$\nand $y[n]$ with 10 dB $E_s/N_0$");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True)
    verify_only_one_specified(snr, noise)

    if snr is not None:
        verify_scalar(snr, float=True)
        snr_linear = linear(snr)
        signal_power = average_power(x)
        noise_power = signal_power / snr_linear
    elif noise is not None:
        verify_scalar(noise, float=True, non_negative=True)
        noise_power = noise

    rng = np.random.default_rng(seed)
    if np.iscomplexobj(x):
        w = rng.normal(0, np.sqrt(noise_power / 2), x.shape) + 1j * rng.normal(0, np.sqrt(noise_power / 2), x.shape)
    else:
        w = rng.normal(0, np.sqrt(noise_power), x.shape)

    y = x + w

    return convert_output(y)


@export
def iq_imbalance(x: npt.ArrayLike, amplitude: float, phase: float = 0.0) -> npt.NDArray[np.complex128]:
    r"""
    Applies IQ imbalance to the complex time-domain signal $x[n]$.

    Arguments:
        x: The complex time-domain signal $x[n]$ to which IQ imbalance is applied.
        amplitude: The amplitude imbalance $A$ in dB. A positive value indicates that the in-phase component is
            larger than the quadrature component.
        phase: The phase imbalance $\phi$ in degrees. A positive value indicates that the quadrature component
            leads the in-phase component.

    Returns:
        The signal $x[n]$ with IQ imbalance applied.

    Notes:
        The IQ imbalance is applied as follows.

        $$g_I = 10^{(A/2)/20} \exp\left(j \frac{-\phi}{2} \frac{\pi}{180}\right)$$
        $$g_Q = 10^{(-A/2)/20} \exp\left(j \frac{\phi}{2} \frac{\pi}{180}\right)$$
        $$y[n] = g_I \cdot x_I[n] + j \cdot g_Q \cdot x_Q[n]$$

    Examples:
        Positive amplitude imbalance horizontally stretches the constellation, while negative amplitude imbalance
        vertically stretches the constellation.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, 4, 1_000); \
            x = psk.map_symbols(s); \
            y1 = sdr.iq_imbalance(x, 5, 0); \
            y2 = sdr.iq_imbalance(x, -5, 0)

            @savefig sdr_iq_imbalance_1.png
            plt.figure(); \
            plt.subplot(1, 2, 1); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y1, label="$y_1[n]$"); \
            plt.title("5 dB amplitude imbalance"); \
            plt.subplot(1, 2, 2); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y2, label="$y_2[n]$"); \
            plt.title("-5 dB amplitude imbalance");

        Positive phase imbalance stretches to the northwest, while negative phase imbalance stretches to the
        northeast

        .. ipython:: python

            y1 = sdr.iq_imbalance(x, 0, 20); \
            y2 = sdr.iq_imbalance(x, 0, -20)

            @savefig sdr_iq_imbalance_2.png
            plt.figure(); \
            plt.subplot(1, 2, 1); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y1, label="$y_1[n]$"); \
            plt.title("20 deg phase imbalance"); \
            plt.subplot(1, 2, 2); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y2, label="$y_2[n]$"); \
            plt.title("-20 deg phase imbalance");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, imaginary=True)
    verify_scalar(amplitude, float=True)
    verify_scalar(phase, float=True)

    phase = np.deg2rad(phase)

    # TODO: Should the phase be negative for I?
    gain_i = linear(0.5 * amplitude, type="voltage") * np.exp(1j * -0.5 * phase)
    gain_q = linear(-0.5 * amplitude, type="voltage") * np.exp(1j * 0.5 * phase)

    y = gain_i * x.real + 1j * gain_q * x.imag

    return convert_output(y)


@export
def sample_rate_offset(
    x: npt.ArrayLike,
    offset: npt.ArrayLike,
    offset_rate: npt.ArrayLike = 0.0,
    sample_rate: float = 1.0,
) -> npt.NDArray:
    r"""
    Applies a sample rate offset to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which the sample rate offset is applied.
        offset: The sample rate offset $\Delta f_s = f_{s,\text{new}} - f_{s}$ in samples/s.
        offset_rate: The sample rate offset rate $\Delta^2 f_s / \Delta t$ in samples/s^2.
        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The signal $x[n]$ with sample rate offset applied.

    Notes:
        The sample rate offset is applied using a Farrow resampler. The resampling rate is calculated as follows.

        $$
        \text{rate} = \frac{f_s + \Delta f_s + \frac{\Delta f_s}{f_s}}{f_s}
        $$

    Examples:
        Create a QPSK reference signal.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, psk.order, 1_000); \
            x = psk.map_symbols(s)

        Add 10 ppm of sample rate offset.

        .. ipython:: python

            y = sdr.sample_rate_offset(x, 10e-6)

            @savefig sdr_sample_rate_offset_1.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title("10 ppm sample rate offset");

        Add 100 ppm of sample rate offset.

        .. ipython:: python

            y = sdr.sample_rate_offset(x, 100e-6)

            @savefig sdr_sample_rate_offset_2.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title("100 ppm sample rate offset");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    offset = verify_arraylike(offset, float=True)
    offset_rate = verify_arraylike(offset_rate, float=True)
    verify_scalar(sample_rate, float=True, positive=True)

    # TODO: Is this correct....
    rate = (sample_rate + offset + offset_rate / sample_rate) / sample_rate
    farrow = FarrowResampler()
    y = farrow(x, rate)

    return convert_output(y)


@export
def frequency_offset(
    x: npt.ArrayLike,
    offset: npt.ArrayLike,
    offset_rate: npt.ArrayLike = 0.0,
    phase: npt.ArrayLike = 0.0,
    sample_rate: float = 1.0,
) -> npt.NDArray:
    r"""
    Applies a frequency and phase offset to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which the frequency offset is applied.
        offset: The frequency offset $\Delta f = f_{\text{new}} - f$ in Hz.
        offset_rate: The frequency offset rate $\Delta^2 f / \Delta t$ in Hz/s. For example, a frequency
            offset varying from 1 kHz to 2 kHz over 1 ms, the offset rate is 1 kHz / 1 ms or 1 MHz/s.
        phase: The phase offset $\phi$ in degrees.
        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The signal $x[n]$ with frequency offset applied.

    Examples:
        Create a reference signal with a constant frequency of 1 cycle per 100 samples.

        .. ipython:: python

            x = sdr.sinusoid(100, freq=1 / 100)

        Add a frequency offset of 1 cycle per 100 samples (the length of the signal). Notice that the signal now
        rotates through 2 cycles instead of 1.

        .. ipython:: python

            freq = 1 / 100
            y = sdr.frequency_offset(x, freq)

            @savefig sdr_frequency_offset_1.png
            plt.figure(); \
            sdr.plot.time_domain(np.unwrap(np.angle(x)) / (2 * np.pi), label="$x[n]$"); \
            sdr.plot.time_domain(np.unwrap(np.angle(y)) / (2 * np.pi), label="$y[n]$"); \
            plt.ylabel("Absolute phase (cycles)"); \
            plt.title("Constant frequency offset (linear phase)");

        Add a frequency rate of change of 2 cycles per 100^2 samples. Notice that the signal now rotates through
        4 cycles instead of 2.

        .. ipython:: python

            freq_rate = 2 / 100**2
            y = sdr.frequency_offset(x, freq, freq_rate)

            @savefig sdr_frequency_offset_2.png
            plt.figure(); \
            sdr.plot.time_domain(np.unwrap(np.angle(x)) / (2 * np.pi), label="$x[n]$"); \
            sdr.plot.time_domain(np.unwrap(np.angle(y)) / (2 * np.pi), label="$y[n]$"); \
            plt.ylabel("Absolute phase (cycles)"); \
            plt.title("Linear frequency offset (quadratic phase)");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    offset = verify_arraylike(offset, float=True)
    offset_rate = verify_arraylike(offset_rate, float=True)
    phase = verify_arraylike(phase, float=True)
    verify_scalar(sample_rate, float=True, positive=True)

    t = np.arange(x.size) / sample_rate  # Time vector in seconds
    f = offset + offset_rate * t  # Frequency vector in Hz
    lo = np.exp(1j * (2 * np.pi * f * t + np.deg2rad(phase)))  # Local oscillator
    y = x * lo  # Apply frequency offset

    return convert_output(y)


@export
def clock_error(
    x: npt.ArrayLike,
    error: npt.ArrayLike,
    error_rate: float = 0.0,
    center_freq: float | None = None,
    sample_rate: float | None = None,
) -> npt.NDArray:
    r"""
    Applies a clock error to the time-domain signal $x[n]$.

    This clock error could be caused by transmitter clock error, receiver clock error, or Doppler effects.

    Arguments:
        x: The time-domain signal $x[n]$ to which the clock error is applied.

            .. warning::

                The signal must be a real passband signal or a complex baseband signal (with 0 Hz baseband frequency).

                If the signal is a real passband signal, time will be compressed resulting in a carrier frequency
                change. If the signal is a complex baseband signal, time will similarly be compressed. However,
                the zero-IF baseband signal will not observe a frequency shift, since it was always mixed to baseband.
                Therefore, there is a subsequent frequency shift corresponding to the expected frequency shift at
                passband.

                If a complex low-IF signal is provided, the IF frequency will be shifted during time compression.
                This can become noticeable at high clock errors, e.g. 1,000 ppm or more. It is not advised to use
                this function with complex low-IF signals.

        error: The fractional clock error $\epsilon$, which is unitless, with 0 representing no clock error. For example,
            1e-6 represents 1 ppm of clock error.

            The fractional clock error can be calculated from frequency offset $\Delta f$ and
            carrier frequency $f_c$ as $\epsilon = \Delta f / f_c$. For example, a 1 kHz frequency error applied to a
            signal with a 1 GHz carrier frequency is 1e-6 or 1 ppm.

            The fractional clock error can also be calculated from sample rate offset $\Delta f_s$ and
            sample rate $f_s$ as $\epsilon = \Delta f_s / f_s$. For example, a 10 Hz sample rate error applied to a
            signal with a 1 MHz sample rate is 1e-5 or 10 ppm.

            The fractional clock error can also be calculated from relative velocity $\Delta v$ and speed of light
            $c$ as $\epsilon = \Delta v / c$. For example, a 60 mph (or 26.82 m/s) relative velocity between the
            transmitter and receiver is 8.946e-8 or 8.9 ppb.

        error_rate: The clock error $\Delta \epsilon / \Delta t$ in 1/s.
        center_freq: The center frequency $f_c$ of the complex baseband signal in Hz. 0 Hz baseband frequency must
            correspond to the signal's carrier frequency. If $x[n]$ is complex, this must be provided.
        sample_rate: The sample rate $f_s$ in samples/s. If $x[n]$ is complex, this must be provided.

    Returns:
        The signal $x[n]$ with clock error applied.

    Examples:
        This example demonstrates the effect of clock error on a real passband signal. The signal has a carrier
        frequency of 100 kHz. A frequency offset of 20 kHz is desired, corresponding to a clock error or 0.2.
        The clock error is added to the transmitter, and then removed at the receiver. Notice that the transmitted
        signal is compressed in time and shifted in frequency. Also notice that the corrected received signal
        matches the original.

        .. ipython:: python

            sample_rate = 2e6; \
            freq = 100e3; \
            duration = 1000e-6; \
            x = sdr.sinusoid(duration, freq, sample_rate=sample_rate, complex=False)

            freq_offset = 20e3; \
            error = freq_offset / freq; \
            print("Clock error:", error); \
            y = sdr.clock_error(x, error)

            error = -error / (1 + error); \
            print("Clock error:", error); \
            z = sdr.clock_error(y, error)

            @savefig sdr_clock_error_1.png
            plt.figure(); \
            sdr.plot.time_domain(x - 0, sample_rate=sample_rate, label="No clock error"); \
            sdr.plot.time_domain(y - 3, sample_rate=sample_rate, label="Added Tx clock error"); \
            sdr.plot.time_domain(z - 6, sample_rate=sample_rate, label="Removed Tx clock error"); \
            plt.legend(loc="lower left"); \
            plt.title("Real passband signals with and without clock error");

            @savefig sdr_clock_error_2.png
            plt.figure(); \
            sdr.plot.dtft(x, sample_rate=sample_rate, label="No clock error"); \
            sdr.plot.dtft(y, sample_rate=sample_rate, label="Added Tx clock error"); \
            sdr.plot.dtft(z, sample_rate=sample_rate, label="Removed Tx clock error"); \
            plt.axvline(freq, color="k", linestyle="--"); \
            plt.axvline(freq + freq_offset, color="k", linestyle="--"); \
            plt.xlim(80e3, 140e3);

        This example demonstrates the effect of clock error on a complex baseband signal. The signal has a carrier
        frequency of 1 MHz and sample rate of 2 MS/s. A frequency offset of 100 kHz is desired, corresponding to a
        clock error of 0.1. The clock error is added to the transmitter, and then removed at the receiver. Notice
        that the transmitted signal is compressed in time, but not shifted in frequency. Notice that the transmitted
        signal is compressed in time and shifted in frequency. Also notice that the corrected received signal
        matches the original.

        .. ipython:: python

            sample_rate = 2e6; \
            center_freq = 1e6; \
            duration = 1000e-6; \
            x = sdr.sinusoid(duration, 0, sample_rate=sample_rate)

            freq_offset = 100e3; \
            error = freq_offset / center_freq; \
            print("Clock error:", error); \
            y = sdr.clock_error(x, error, 0, center_freq, sample_rate=sample_rate)

            error = -error / (1 + error); \
            print("Clock error:", error); \
            z = sdr.clock_error(y, error, 0, center_freq, sample_rate=sample_rate)

            @savefig sdr_clock_error_3.png
            plt.figure(); \
            sdr.plot.time_domain(x - 0 - 0j, sample_rate=sample_rate, label="No clock error"); \
            sdr.plot.time_domain(y - 3 - 3j, sample_rate=sample_rate, label="Added Tx clock error"); \
            sdr.plot.time_domain(z - 6 - 6j, sample_rate=sample_rate, label="Removed Tx clock error"); \
            plt.legend(loc="lower left"); \
            plt.title("Complex baseband signals with and without clock error");

            @savefig sdr_clock_error_4.png
            plt.figure(); \
            sdr.plot.dtft(x, sample_rate=sample_rate, label="No clock error"); \
            sdr.plot.dtft(y, sample_rate=sample_rate, label="Added Tx clock error"); \
            sdr.plot.dtft(z, sample_rate=sample_rate, label="Removed Tx clock error"); \
            plt.axvline(0, color="k", linestyle="--"); \
            plt.axvline(freq_offset, color="k", linestyle="--"); \
            plt.xlim(-20e3, 120e3);

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    error = verify_arraylike(error, float=True)
    verify_scalar(error_rate, float=True)

    # Apply time compression using resampling
    alpha = 1 + error
    # y = sample_rate_offset(x, 1 / alpha, 0)  # TODO: This doesn't work...
    farrow = FarrowResampler()
    y = farrow(x, 1 / alpha)

    if np.issubdtype(x.dtype, np.floating):
        verify_not_specified(center_freq)

        # The carrier frequency was already shifted by the time compression
        z = y
    else:
        verify_scalar(center_freq, float=True, positive=True)
        verify_scalar(sample_rate, float=True, positive=True)

        # Apply frequency shift that would be observed at passband
        freq_offset = error * center_freq  # Hz
        freq_offset_rate = error_rate * center_freq  # Hz/s
        z = frequency_offset(y, freq_offset, freq_offset_rate, sample_rate=sample_rate)

    return convert_output(z)



import numpy as np

import sdr


def test_complex():
    sample_rate = 1e3  # samples/s
    N = 100  # samples
    freq = 10  # Hz
    phase = 45  # degrees
    lo = sdr.sinusoid(N / sample_rate, freq=freq, phase=phase, sample_rate=sample_rate, complex=True)
    lo_truth = np.exp(1j * (2 * np.pi * freq / sample_rate * np.arange(N) + np.deg2rad(phase)))
    assert np.allclose(lo, lo_truth)


def test_real():
    sample_rate = 1e3  # samples/s
    N = 100  # samples
    freq = 10  # Hz
    phase = 45  # degrees
    lo = sdr.sinusoid(N / sample_rate, freq=freq, phase=phase, sample_rate=sample_rate, complex=False)
    lo_truth = np.cos(2 * np.pi * freq / sample_rate * np.arange(N) + np.deg2rad(phase))
    assert np.allclose(lo, lo_truth)
